# Part 5: AI 핵심 라이브러리 용어집

**[⬆️ 파트 5으로 돌아가기](./part_5_ai_core_libraries.md)**

---

## 용어 목록 (Glossary)

- [NumPy](#numpy)
- [ndarray](#ndarray)
- [Pandas](#pandas)
- [DataFrame](#dataframe)
- [Series](#series)
- [벡터화 (Vectorization)](#벡터화-vectorization)
- [브로드캐스팅 (Broadcasting)](#브로드캐스팅-broadcasting)
- [인덱싱 (Indexing) & 필터링 (Filtering)](#인덱싱-indexing--필터링-filtering)
- [loc](#loc)
- [iloc](#iloc)
- [groupby](#groupby)
- [merge](#merge)
- [concat](#concat)
- [pivot_table](#pivot_table)

---

### **NumPy**

- **설명**: "Numerical Python"의 약자로, 파이썬에서 과학 및 수치 계산을 위한 가장 기본적인 핵심 라이브러리입니다. 고성능 다차원 배열 객체인 `ndarray`를 중심으로 벡터, 행렬 등 고성능 연산을 위한 방대한 함수를 제공합니다.

### **ndarray**

- **설명**: NumPy의 핵심 자료구조인 N-dimensional array(다차원 배열)입니다. 파이썬 리스트와 달리 동일한 자료형의 데이터만 담을 수 있어 메모리 효율이 높고, '계란판'처럼 구조가 균일하여 매우 빠른 속도로 원소 전체에 대한 연산(벡터화)이 가능합니다.

### **Pandas**

- **설명**: 행과 열로 이루어진 표 형식의 데이터(Tabular Data)를 쉽게 다룰 수 있게 해주는 데이터 분석 라이브러리입니다. `DataFrame`과 `Series`라는 강력한 자료구조를 제공하며, 데이터 정제, 처리, 분석에 필수적인 도구입니다. 기능이 강력하여 '슈퍼 엑셀'에 비유되기도 합니다.

### **DataFrame**

- **설명**: Pandas의 핵심 자료구조로, 2차원 테이블(표) 형태의 데이터를 담습니다. 행(row)과 열(column)으로 구성되며, 각 열은 서로 다른 데이터 타입을 가질 수 있습니다. 엑셀 스프레드시트나 데이터베이스 테이블과 유사한 구조입니다.

### **Series**

- **설명**: Pandas의 1차원 배열 형태의 자료구조입니다. `DataFrame`의 한 열(column)이 하나의 `Series`에 해당합니다. 데이터 배열과 함께 인덱스(index)를 가집니다.

### **벡터화 (Vectorization)**

- **설명**: 반복문을 사용하지 않고 배열의 모든 요소에 대해 한 번에 연산을 수행하는 방식입니다. NumPy가 고속으로 동작하는 핵심 원리이며, 코드를 간결하고 효율적으로 만듭니다.
- **예시**: `for` 문으로 배열의 각 원소에 2를 곱하는 대신, `numpy_array * 2` 와 같이 한 줄로 연산을 처리합니다.

### **브로드캐스팅 (Broadcasting)**

- **설명**: NumPy에서 서로 다른 크기(shape)를 가진 배열 간에도 산술 연산이 가능하도록, 특정 조건 하에 배열을 자동으로 확장하여 크기를 맞춰주는 기능입니다.
- **예시**: (3, 3) 모양의 행렬과 (3,) 모양의 벡터를 더할 때, 벡터가 행렬의 각 행에 더해지도록 자동으로 확장됩니다.

### **인덱싱 (Indexing) & 필터링 (Filtering)**

- **설명**: `DataFrame`이나 `ndarray`에서 특정 위치의 데이터를 선택하거나, 특정 조건을 만족하는 데이터 부분집합을 추출하는 작업을 의미합니다.
- **예시**: `df['age'] > 30` 와 같이 특정 조건을 만족하는 행만 걸러내는 것이 필터링입니다.

### **loc**

- **설명**: "Location"의 약자로, Pandas `DataFrame`에서 **레이블(이름)** 기반으로 데이터를 선택(indexing)하는 속성입니다. 행과 열의 이름을 직접 지정하여 데이터를 추출합니다.
- **사용법**: `df.loc[행_이름, 열_이름]`

### **iloc**

- **설명**: "Integer Location"의 약자로, Pandas `DataFrame`에서 **정수 위치(0부터 시작하는 순서)** 기반으로 데이터를 선택하는 속성입니다.
- **사용법**: `df.iloc[행_번호, 열_번호]`

### **groupby**

- **설명**: 데이터를 특정 기준으로 그룹으로 묶고(Split), 각 그룹에 특정 함수(예: `sum`, `mean`)를 적용(Apply)한 뒤, 그 결과를 하나로 합쳐서(Combine) 보여주는 기능입니다. '반별 평균 성적 계산'과 같은 집계 분석에 매우 유용합니다.

### **merge**

- **설명**: 두 `DataFrame`을 특정 공통 열(Key)을 기준으로 합치는 기능입니다. 데이터베이스의 `JOIN` 연산과 매우 유사하며, 여러 데이터 소스를 하나로 결합할 때 사용됩니다.
- **예시**: '학생 정보' 테이블과 '성적 정보' 테이블을 '학번'을 기준으로 합칠 수 있습니다.

### **concat**

- **설명**: "Concatenate"의 약자로, 여러 `DataFrame`을 행(row)이나 열(column) 방향으로 그대로 이어 붙이는 기능입니다. `merge`처럼 특정 키를 기준으로 맞추는 것이 아니라, 단순히 아래나 옆으로 데이터를 연결합니다.

### **pivot_table**

- **설명**: 엑셀의 피벗 테이블 기능처럼, `DataFrame`의 데이터를 재구성하여 새로운 관점의 요약 테이블을 만드는 기능입니다. 특정 열을 행과 열 인덱스로 사용하고, 다른 열의 데이터로 값을 채워 데이터를 요약, 분석할 수 있습니다. 