# Part 10: 전문가로 가는 길 용어집

## 핵심 용어

| 용어 (Term) | 설명 (Description) |
| --- | --- |
| **MLOps (Machine Learning Operations)** | 머신러닝(ML) 모델 개발(Dev)과 운영(Ops)을 통합하여, ML 시스템의 개발, 배포, 유지보수를 자동화하고 효율화하는 일련의 원칙과 실천 방법입니다. CI/CD, 데이터/모델 버전 관리, 모니터링, 재학습 등을 포함하여 AI 서비스의 안정성과 지속성을 높입니다. (비유: AI 레스토랑 프랜차이즈의 '중앙 관제 시스템') |
| **CI/CD (Continuous Integration/Continuous Deployment or Delivery)** | 지속적 통합(CI)은 코드 변경사항을 정기적으로 중앙 리포지토리에 병합하고, 자동으로 빌드 및 테스트하는 프로세스입니다. 지속적 배포/딜리버리(CD)는 CI를 통과한 코드를 자동으로 프로덕션 환경까지 배포하거나, 배포 가능한 상태로 준비하는 프로세스입니다. MLOps에서 모델 및 애플리케이션의 빠르고 안정적인 업데이트를 가능하게 합니다. |
| **DVC (Data Version Control)** | Git과 함께 사용되어 대용량 데이터셋, 머신러닝 모델, 중간 파일 등의 버전을 관리하는 오픈소스 도구입니다. 실제 데이터 파일 대신 메타데이터(포인터)를 Git으로 관리하여, 대용량 파일로 인해 Git 리포지토리가 비대해지는 것을 방지하고 실험의 재현성을 높입니다. |
| **MLflow** | 머신러닝 수명주기 관리를 위한 오픈소스 플랫폼입니다. 실험 추적(파라미터, 코드, 데이터, 결과 로깅), 모델 패키징 및 버전 관리, 모델 배포 기능을 제공하여, ML 실험의 재현성을 높이고 협업을 용이하게 합니다. |
| **AI 네이티브 IDE (AI Native IDE)** | 인공지능 기능이 깊숙하게 통합된 통합 개발 환경(IDE)입니다. 코드 자동 생성, 리팩터링, 디버깅, 문서화 등 개발 과정 전반에 걸쳐 AI의 지원을 받아 생산성을 극대화합니다. (예: Cursor, GitHub Copilot이 통합된 VS Code) (비유: AI 비서(자비스)가 탑재된 F-35 스텔스기 조종석) |
| **Cursor** | AI 기능을 중심으로 설계된 코드 편집기로, GPT-4와 같은 LLM을 활용하여 코드 작성, 수정, 채팅 기반의 코드 관련 질문 답변, 버그 수정, 문서 참조 등 다양한 개발 작업을 지원하는 AI 네이티브 IDE의 한 종류입니다. |
| **GitHub Copilot** | GitHub와 OpenAI가 개발한 AI 페어 프로그래머(pair programmer) 서비스입니다. 개발자가 코드를 작성할 때 실시간으로 코드 라인이나 전체 함수를 제안하며, 주석을 코드로 변환하거나 테스트 코드를 생성하는 등 개발 생산성을 높이는 데 도움을 줍니다. VS Code 등 다양한 IDE의 확장 기능으로 제공됩니다. |
| **SLM (Small Language Model / 소형 언어 모델)** | 거대 언어 모델(LLM)의 성능을 특정 작업이나 도메인에서 유지하면서, 모델 크기와 계산 요구량을 크게 줄인 효율적인 언어 모델입니다. 스마트폰, 노트북 등 온디바이스(On-device) 환경에서의 실행을 목표로 하며, 속도, 비용, 개인정보 보호 측면에서 이점을 가집니다. (예: Llama 3 8B, Phi-3) |
| **멀티모달 AI (Multimodal AI)** | 텍스트, 이미지, 음성, 비디오 등 여러 종류의 데이터(양식, modality)를 동시에 이해하고 처리하며 생성할 수 있는 인공지능입니다. 인간처럼 다양한 감각 정보를 통합적으로 활용하여 보다 복잡하고 풍부한 상호작용 및 문제 해결이 가능합니다. (예: GPT-4o, Google Gemini) |
| **AI 에이전트 (AI Agent)** | 특정 목표를 달성하기 위해 자율적으로 환경을 관찰하고, 계획을 세우며, 도구(API, 코드 실행 등)를 사용하고, 행동을 실행하는 AI 시스템입니다. 단순 질의응답을 넘어, 복잡한 실제 태스크를 사용자를 대신하여 수행할 수 있는 잠재력을 가집니다. (예: Auto-GPT, LangGraph 기반 에이전트) |
| **AI 윤리 (AI Ethics)** | 인공지능 기술의 개발 및 사용과 관련된 윤리적 문제들을 다루는 분야입니다. 데이터 편향성, 개인정보 침해, 알고리즘의 공정성 및 투명성, 책임성, 일자리 대체, 잠재적 오용 등 AI가 사회에 미치는 영향을 고려하고 바람직한 방향으로 기술을 발전시키기 위한 원칙과 가이드라인을 포함합니다. |

---

*이 용어집은 [`part_10_expert_path.md`](./part_10_expert_path.md) 파일 내용을 기반으로 작성되었습니다.* 